import os
import signal
import sys
import time
import socket
import subprocess
import psutil
from contextlib import closing
from typing import Optional

class ServerManager:
    def __init__(self, app_path: str, host: str, port: int, name: str):
        self.app_path = app_path
        self.host = host
        self.port = port
        self.name = name
        self._process: Optional[subprocess.Popen] = None
        self._pid_file = os.path.join(os.path.expanduser("~"), f".upsonic_{name}_server.pid")

    def _is_port_in_use(self) -> bool:
        """Check if the port is in use."""
        try:
            # Faster method to check port availability
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(0.2)  # Reduce timeout for faster checks
            result = sock.connect_ex((self.host, self.port))
            sock.close()
            return result == 0
        except Exception:
            # Fallback to the original method
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                return sock.connect_ex((self.host, self.port)) == 0

    def _kill_process_using_port(self) -> bool:
        """Find and kill processes using the specified port."""
        killed = False
        
        for proc in psutil.process_iter(['pid', 'name']):
            try:
                process = psutil.Process(proc.info['pid'])
                for conn in process.net_connections():
                    if hasattr(conn, 'laddr') and len(conn.laddr) >= 2 and conn.laddr[1] == self.port:
                        try:
                            # Kill process more aggressively
                            process.kill()
                            process.wait(timeout=0.5)  # Reduced timeout
                            killed = True
                        except Exception:
                            # Last resort: try system kill
                            try:
                                os.kill(process.pid, signal.SIGKILL)
                                killed = True
                            except Exception:
                                pass
            except Exception:
                continue
                
        return killed

    def _manage_pid_file(self, operation: str):
        """Manage PID file operations (read/write/cleanup)."""
        try:
            if operation == "write" and self._process and self._process.pid:
                with open(self._pid_file, 'w') as f:
                    f.write(str(self._process.pid))
            elif operation == "read" and os.path.exists(self._pid_file):
                with open(self._pid_file, 'r') as f:
                    return int(f.read().strip())
            elif operation == "cleanup" and os.path.exists(self._pid_file):
                os.remove(self._pid_file)
        except Exception:
            pass
        
        return None if operation == "read" else False

    def _cleanup_port(self):
        """Clean up the port before starting."""
        if not self._is_port_in_use():
            return True
            
        # Try to kill processes using the port
        self._kill_process_using_port()
        time.sleep(0.2)  # Reduced sleep time
        
        # If port is still in use, try one more aggressive approach
        if self._is_port_in_use():
            try:
                # Try to bind to the port to force it free
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind((self.host, self.port))
                sock.close()
            except socket.error:
                # If binding fails, try one more kill attempt
                for proc in psutil.process_iter(['pid']):
                    try:
                        process = psutil.Process(proc.info['pid'])
                        if any(len(conn.laddr) >= 2 and conn.laddr[1] == self.port 
                              for conn in process.net_connections(kind='inet')):
                            process.kill()
                    except Exception:
                        continue
                time.sleep(0.2)  # Reduced sleep time
        
        return not self._is_port_in_use()

    def start(self, redirect_output: bool = False, force: bool = False):
        """Start the server if it's not already running."""
        if self.is_running():
            return
            
        # Clean up port
        if not self._cleanup_port() and not force:
            raise RuntimeError(f"Port {self.port} is in use and could not be freed")

        # Set up logging
        stdout = stderr = None
        if redirect_output:
            log_dir = "logs"
            os.makedirs(log_dir, exist_ok=True)
            stdout = open(os.path.join(log_dir, f'{self.name}_server.log'), 'a')
            stderr = open(os.path.join(log_dir, f'{self.name}_server_error.log'), 'a')

        # Start the server process
        try:
            cmd = [
                sys.executable, "-m", "uvicorn",
                self.app_path,
                "--host", self.host,
                "--port", str(self.port),
                "--log-level", "error",
                "--no-access-log"
            ]
            
            self._process = subprocess.Popen(
                cmd, stdout=stdout, stderr=stderr, start_new_session=True
            )
            self._manage_pid_file("write")

            # Wait for server to start with optimized polling
            # Initial quick sleep to allow process to start
            time.sleep(0.1)
            
            # Progressive polling with increasing intervals
            poll_interval = 0.01
            max_poll_interval = 0.1
            start_time = time.time()
            while not self._is_port_in_use() and time.time() - start_time < 30:
                if self._process.poll() is not None:
                    raise RuntimeError(f"Server process terminated unexpectedly with code {self._process.returncode}")
                time.sleep(poll_interval)
                # Gradually increase polling interval
                poll_interval = min(poll_interval * 1.2, max_poll_interval)

            if not self._is_port_in_use():
                raise RuntimeError(f"Timeout waiting for {self.name} server to start")
                
        except Exception as e:
            self.stop()
            raise RuntimeError(f"Failed to start {self.name} server: {str(e)}")

    def stop(self):
        """Stop the server if it's running."""
        # Try to stop process from PID file
        pid = self._manage_pid_file("read")
        if pid:
            try:
                process = psutil.Process(pid)
                try:
                    os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                    process.wait(timeout=5)
                except Exception:
                    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
            except Exception:
                pass

        # Try to stop process from instance variable
        if self._process:
            try:
                self._process.terminate()
                self._process.wait(timeout=5)
            except Exception:
                if self._process.poll() is None:
                    self._process.kill()

        self._process = None
        self._manage_pid_file("cleanup")

    def is_running(self) -> bool:
        """Check if the server is currently running."""
        # Check if process from instance is running
        if self._process and self._process.poll() is None:
            return True

        # Check if process from PID file is running
        pid = self._manage_pid_file("read")
        if pid:
            try:
                process = psutil.Process(pid)
                return process.is_running() and process.name().startswith("python")
            except psutil.NoSuchProcess:
                self._manage_pid_file("cleanup")
                
        return False 


import copy
import time
import cloudpickle
import asyncio

from ..knowledge_base.knowledge_base import KnowledgeBase
cloudpickle.DEFAULT_PROTOCOL = 2

import dill
import base64
import httpx
from typing import Any, List, Dict, Optional, Type, Union
from pydantic import BaseModel

from ..tasks.tasks import Task

from ..printing import call_end


from ..tasks.task_response import ObjectResponse


from ..level_utilized.utility import context_serializer, response_format_serializer, tools_serializer, response_format_deserializer, error_handler

class Call:


    def call(
        self,
        task: Union[Task, List[Task]],
        llm_model: str = None,
    ) -> Any:
        
        start_time = time.time()

        try:
            # Try to get the current event loop
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    # If there's a running loop, run the async function in that loop
                    if isinstance(task, list):
                        for each in task:
                            the_result = asyncio.run_coroutine_threadsafe(self.call_async_(each, llm_model), loop).result()
                            call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
                    else:
                        the_result = asyncio.run_coroutine_threadsafe(self.call_async_(task, llm_model), loop).result()
                        call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
                else:
                    # If there's a loop but it's not running, use asyncio.run
                    if isinstance(task, list):
                        for each in task:
                            the_result = asyncio.run(self.call_async_(each, llm_model))
                            call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
                    else:
                        the_result = asyncio.run(self.call_async_(task, llm_model))
                        call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
            except RuntimeError:
                # No event loop exists, create one with asyncio.run
                if isinstance(task, list):
                    for each in task:
                        the_result = asyncio.run(self.call_async_(each, llm_model))
                        call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
                else:
                    the_result = asyncio.run(self.call_async_(task, llm_model))
                    call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
        except Exception as outer_e:
            try:
                from ...server import stop_dev_server, stop_main_server, is_tools_server_running, is_main_server_running

                if is_tools_server_running() or is_main_server_running():
                    stop_dev_server()

            except Exception:
                pass

            raise outer_e

        end_time = time.time()

        return task.response

    def call_(
        self,
        task: Task,
        llm_model: str = None,
    ) -> Any:
        """
        Call GPT-4 with optional tools and MCP servers.

        Args:
            prompt: The input prompt for GPT-4
            response_format: The expected response format (can be a type or Pydantic model)
            tools: Optional list of tool names to use


        Returns:
            The response in the specified format
        """
        # Try to get the current event loop
        try:
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the async function in that loop
                return asyncio.run_coroutine_threadsafe(self.call_async_(task, llm_model), loop).result()
            else:
                # If there's a loop but it's not running, use asyncio.run
                return asyncio.run(self.call_async_(task, llm_model))
        except RuntimeError:
            # No event loop exists, create one with asyncio.run
            return asyncio.run(self.call_async_(task, llm_model))

    async def call_async(
        self,
        task: Union[Task, List[Task]],
        llm_model: str = None,
    ) -> Any:
        """
        Asynchronous version of the call method.
        """
        start_time = time.time()

        try:
            if isinstance(task, list):
                for each in task:
                    the_result = await self.call_async_(each, llm_model)
                    call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
            else:
                the_result = await self.call_async_(task, llm_model)
                call_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], self.debug)
        except Exception as outer_e:
            try:
                from ...server import stop_dev_server, stop_main_server, is_tools_server_running, is_main_server_running

                if is_tools_server_running() or is_main_server_running():
                    stop_dev_server()
            except Exception:
                pass
            raise outer_e

        end_time = time.time()

        return task.response

    async def call_async_(
        self,
        task: Task,
        llm_model: str = None,
    ) -> Any:
        """
        Asynchronous version of the call_ method.
        """
        task.start_time = time.time()
        from ..trace import sentry_sdk
        
        # Use the provided model or default to the client's default
        if llm_model is None:
            llm_model = self.default_llm_model
            
        tools = tools_serializer(task.tools)

        response_format = task.response_format
        with sentry_sdk.start_transaction(op="task", name="Call.call_async") as transaction:
            with sentry_sdk.start_span(op="serialize"):
                # Serialize the response format if it's a type or BaseModel
                response_format_str = response_format_serializer(task.response_format)

                new_context = []
                if task.context:
                    for each in task.context:
                        if isinstance(each, KnowledgeBase):
                            if not each.rag:
                                new_context.append(each.markdown(self))
                        else:
                            new_context.append(each)

                context = context_serializer(new_context, self)

            with sentry_sdk.start_span(op="prepare_request"):
                # Prepare the request data
                data = {
                    "prompt": task.description + task.additional_description(self), 
                    "images": task.images_base_64,
                    "response_format": response_format_str,
                    "tools": tools or [],
                    "context": context,
                    "llm_model": llm_model,
                    "system_prompt": None
                }

            with sentry_sdk.start_span(op="send_request"):
                result = await self.send_request_async("/level_one/gpt4o", data)
                original_result = result
                
                result = result["result"]
                
                error_handler(result)

            with sentry_sdk.start_span(op="deserialize"):
                deserialized_result = response_format_deserializer(response_format_str, result)

        task._response = deserialized_result["result"]
        
        response_format_req = None
        if response_format_str == "str":
            response_format_req = response_format_str
        else:
            # Class name
            response_format_req = response_format.__name__
        
        task.end_time = time.time()
        
        return {"result": deserialized_result["result"], "llm_model": llm_model, "response_format": response_format_req, "usage": deserialized_result["usage"]}






from typing import List
from pydantic import BaseModel


class ObjectResponse(BaseModel):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._upsonic_response_type = "default"

class CustomTaskResponse(BaseModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._upsonic_response_type = "custom"

    def output(self):
        # Get the first field name from the model
        field_name = next(iter(self.__annotations__))
        # Return the value of that field
        return getattr(self, field_name)




def IntResponse(name: str):

   
    name = name.lower().replace(" ", "_")

    
    class IntegerResponse(CustomTaskResponse):
        __annotations__ = {name: int}
        locals()[name] = None
    IntegerResponse.__name__ = name.capitalize()
    return IntegerResponse


def FloatResponse(name: str):

    
    name = name.lower().replace(" ", "_")
    class FloatingResponse(CustomTaskResponse):
        __annotations__ = {name: float}
        locals()[name] = None
    FloatingResponse.__name__ = name.capitalize()
    return FloatingResponse


def BoolResponse(name: str):

    
    name = name.lower().replace(" ", "_")   
    class BooleanResponse(CustomTaskResponse):
        __annotations__ = {name: bool}
        locals()[name] = None
    BooleanResponse.__name__ = name.capitalize()
    return BooleanResponse


def StrResponse(name: str):

    
    name = name.lower().replace(" ", "_")
    class StringResponse(CustomTaskResponse):
        __annotations__ = {name: str}
        locals()[name] = None
    StringResponse.__name__ = name.capitalize()
    return StringResponse

def StrInListResponse(name: str):
    name = name.lower().replace(" ", "_")
    class StrInListResponse(CustomTaskResponse):
        __annotations__ = {name: list[str]}
        locals()[name] = None
    StrInListResponse.__name__ = name.capitalize()
    return StrInListResponse


from upsonic.tools_server.function_client import FunctionToolManager


def test_function_client_basic():
    # Initialize the client
    client = FunctionToolManager()

    # Get list of tools
    tool_list = client.tools()

    # Check if we got any tools
    assert len(tool_list) > 0

    # Test the first tool (add_numbers)
    first_tool = tool_list[0]

    # Check tool attributes
    assert hasattr(first_tool, "__name__")
    assert hasattr(first_tool, "__doc__")
    assert hasattr(first_tool, "__annotations__")

    # Test tool execution
    result = first_tool(1, 2)
    assert isinstance(result, dict)
    assert "result" in result


def test_get_tools_by_name():
    client = FunctionToolManager()

    # Get specific tools by name
    tools = client.get_tools_by_name(["add_numbers"])

    # Check if we got the tool
    assert len(tools) == 1
    assert tools[0].__name__ == "add_numbers"

    # Test the tool
    result = tools[0](5, 3)
    assert isinstance(result, dict)
    assert "result" in result
    assert result["result"] == 8


def test_direct_tool_call():
    client = FunctionToolManager()

    # Test direct tool call
    result = client.call_tool("add_numbers", {"a": 10, "b": 20})
    assert isinstance(result, dict)
    assert "result" in result
    assert result["result"] == 30



import toml
import os

def get_library_version() -> str:
    """
    Get the version of the upsonic library from pyproject.toml.

    Returns:
        The version of the upsonic library as a string.
    """
    try:
        # Assuming the pyproject.toml is located at the root of the project
        pyproject_path = os.path.join(os.path.dirname(__file__), '../../pyproject.toml')
        with open(pyproject_path, 'r') as file:
            pyproject_data = toml.load(file)
            return pyproject_data['project']['version']
    except (FileNotFoundError, KeyError):
        return "Version information not available."



from fastapi import HTTPException, UploadFile, File
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
import traceback
from ...api import app, timeout

import asyncio
from concurrent.futures import ThreadPoolExecutor
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64
from ....storage.configuration import Configuration
import os
import tempfile


from fastapi.responses import FileResponse
import uuid




prefix = "/others"



@app.get(f"{prefix}/take_screenshot")
async def take_screenshot():
    """
    Takes a screenshot using pyautogui and returns it to the client.

    Returns:
        The screenshot image file
    """
    import pyautogui
    try:
        # Create a temporary directory if it doesn't exist
        temp_dir = os.path.join(tempfile.gettempdir(), "upsonic_screenshots")
        os.makedirs(temp_dir, exist_ok=True)

        # Generate a unique filename
        filename = f"screenshot_{uuid.uuid4()}.png"
        file_path = os.path.join(temp_dir, filename)

        # Take the screenshot
        screenshot = pyautogui.screenshot()
        screenshot.save(file_path)

        # Return the file and clean up after sending
        return FileResponse(
            file_path,
            media_type="image/png",
            filename=filename,
            background=asyncio.create_task(cleanup_screenshot(file_path))
        )
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))

async def cleanup_screenshot(file_path: str):
    """
    Cleanup function to remove the screenshot file after it's been sent.
    """
    try:
        await asyncio.sleep(1)  # Wait a bit to ensure the file has been sent
        if os.path.exists(file_path):
            os.remove(file_path)
    except Exception as e:
        print(f"Error cleaning up screenshot: {e}")



import base64
from pydantic import BaseModel


from typing import Any, List, Dict, Optional, Type, Union


from .task_response import CustomTaskResponse, ObjectResponse
from ..printing import get_price_id_total_cost

from ..knowledge_base.knowledge_base import KnowledgeBase

class Task(BaseModel):
    description: str
    images: Optional[List[str]] = None
    tools: list[Any] = []
    response_format: Union[Type[CustomTaskResponse], Type[ObjectResponse], None] = None
    _response: Any = None
    context: Any = None
    price_id_: str = None
    not_main_task: bool = False
    start_time: Optional[int] = None
    end_time: Optional[int] = None

    @property
    def duration(self) -> Optional[float]:
        if self.start_time is None or self.end_time is None:
            return None
        return self.end_time - self.start_time
    
    def __init__(self, description: str = None, **data):
        if description is not None:
            data["description"] = description
        super().__init__(**data)
        self.validate_tools()

    def validate_tools(self):
        """
        Validates each tool in the tools list.
        If a tool is a class and has a __control__ method, runs that method to verify it returns True.
        Raises an exception if the __control__ method returns False or raises an exception.
        """
        if not self.tools:
            return
            
        for tool in self.tools:
            # Check if the tool is a class
            if isinstance(tool, type) or hasattr(tool, '__class__'):
                # Check if the class has a __control__ method
                if hasattr(tool, '__control__') and callable(getattr(tool, '__control__')):
                    try:
                        # Run the __control__ method
                        control_result = tool.__control__()
                        if not control_result:
                            raise ValueError(f"Tool {tool} __control__ method returned False")
                    except Exception as e:
                        # Re-raise any exceptions from the __control__ method
                        raise ValueError(f"Error validating tool {tool}: {str(e)}")

    
    def additional_description(self, client):
        if not self.context:
            return ""
        
            
        rag_results = []
        for context in self.context:
            
            if isinstance(context, KnowledgeBase) and context.rag == True:
                context.setup_rag(client)
                rag_results.append(context.query(self.description))
                
        if rag_results:
            return f"The following is the RAG data: <rag>{' '.join(rag_results)}</rag>"
        return ""


    @property
    def images_base_64(self):
        if self.images is None:
            return None
        base_64_images = []
        for image in self.images:
            with open(image, "rb") as image_file:
                base_64_images.append(base64.b64encode(image_file.read()).decode('utf-8'))
        return base_64_images

    @property
    def price_id(self):
        if self.price_id_ is None:
            import uuid
            self.price_id_ = str(uuid.uuid4())
        return self.price_id_

    @property
    def response(self):

        if self._response is None:
            return None

        if type(self._response) == str:
            return self._response



        if self._response._upsonic_response_type == "custom":
            return self._response.output()
        else:
            return self._response



    def get_total_cost(self):
        if self.price_id_ is None:
            return None
        return get_price_id_total_cost(self.price_id)
    
    @property
    def total_cost(self) -> Optional[float]:
        total_task_cost = None
        the_total_cost = self.get_total_cost()
        if the_total_cost:
            if "estimated_cost" in the_total_cost:
                total_task_cost = the_total_cost["estimated_cost"]


        return total_task_cost


from dataclasses import Field
import uuid
from pydantic import BaseModel
import asyncio

from typing import Any, List, Dict, Optional, Type, Union

from ..knowledge_base.knowledge_base import KnowledgeBase
from ..tasks.tasks import Task

from ..latest_upsonic_client import latest_upsonic_client

def register_tools(client, tools):
    """Register tools with the client."""
    if tools is not None:
        for tool in tools:
            # Handle special tool classes from upsonic.client.tools
            if tool.__module__ == 'upsonic.client.tools':

                client.tool()(tool)
                continue
                
            # If tool is a class (not an instance)
            if isinstance(tool, type):
                if hasattr(tool, 'command'):

                    client.mcp()(tool)
                else:

                    client.tool()(tool)
            else:
                # Get all attributes of the tool instance/object
                tool_attrs = dir(tool)
                
                # Filter out special methods and get only callable attributes
                functions = [attr for attr in tool_attrs 
                           if not attr.startswith('__') and callable(getattr(tool, attr))]
                
                if functions:
                    # If the tool has functions, use the tool() decorator
                    print("INSTANCE TOOL REGISTERED:", tool.__class__.__name__)

                    if not isinstance(tool, object):
                        client.tool()(tool.__class__)
                    else:
                        client.tool()(tool)
                else:
                    # If the tool has no functions, use mcp()
                    print("INSTANCE MCP REGISTERED:", tool.__class__.__name__)
                    client.mcp()(tool.__class__)
    return client


def get_or_create_client(debug: bool = False):
    """Get existing client or create a new one."""
    
    global latest_upsonic_client
    
    if latest_upsonic_client is not None:
        # Check if the existing client's status is False
        if not latest_upsonic_client.status():
            from ..base import UpsonicClient
            new_client = UpsonicClient("localserver", debug=debug)
            latest_upsonic_client = new_client
        return latest_upsonic_client
    
    from ..base import UpsonicClient
    the_client = UpsonicClient("localserver", debug=debug)
    latest_upsonic_client = the_client
    return the_client


def execute_task(agent_config, task: Task, debug: bool = False):
    """Execute a task with the given agent configuration."""
    import asyncio
    
    try:
        # Check if there's a running event loop
        loop = asyncio.get_running_loop()
        if loop.is_running():
            # If there's a running loop, run the coroutine in that loop
            return asyncio.run_coroutine_threadsafe(
                execute_task_async(agent_config, task, debug), 
                loop
            ).result()
    except RuntimeError:
        # No running event loop
        pass
    
    # If no running loop or exception occurred, create a new one
    return asyncio.run(execute_task_async(agent_config, task, debug))

async def execute_task_async(agent_config, task: Task, debug: bool = False):
    """Execute a task with the given agent configuration asynchronously using true async methods."""
    global latest_upsonic_client
    
    # If agent has a custom client, use it
    if hasattr(agent_config, 'client') and agent_config.client is not None:
        the_client = agent_config.client
    else:
        # Get or create client using existing process
        the_client = get_or_create_client(debug=debug)
    
    # If task has no tools defined but agent has tools, use the agent's tools
    if not task.tools and hasattr(agent_config, 'tools') and agent_config.tools:
        task.tools = agent_config.tools
    
    # Register tools if needed
    the_client = register_tools(the_client, task.tools)
    
    # Use the async run method directly
    await the_client.run_async(agent_config, task)
    
    return task.response


class AgentConfiguration(BaseModel):
    agent_id_: str = None
    job_title: str
    company_url: str = None
    company_objective: str = None
    name: str = ""
    contact: str = ""
    model: str = "openai/gpt-4o"
    client: Any = None  # Add client parameter
    debug: bool = False
    reliability_layer: Any = None  # Changed to Any to accept any class or instance
    system_prompt: str = None
    tools: List[Any] = []

    def __init__(self, job_title: str = None, client: Any = None, **data):
        if job_title is not None:
            data["job_title"] = job_title
        if client is not None:
            data["client"] = client
               

        super().__init__(**data)
        self.validate_tools() 

    sub_task: bool = True
    reflection: bool = False
    memory: bool = False
    caching: bool = True
    cache_expiry: int = 60 * 60
    knowledge_base: KnowledgeBase = None
    context_compress: bool = False



    def validate_tools(self):
        """
        Validates each tool in the tools list.
        If a tool is a class and has a __control__ method, runs that method to verify it returns True.
        Raises an exception if the __control__ method returns False or raises an exception.
        """
        if not self.tools:
            return
            
        for tool in self.tools:
            # Check if the tool is a class
            if isinstance(tool, type) or hasattr(tool, '__class__'):
                # Check if the class has a __control__ method
                if hasattr(tool, '__control__') and callable(getattr(tool, '__control__')):
                    try:
                        # Run the __control__ method
                        control_result = tool.__control__()
                        if not control_result:
                            raise ValueError(f"Tool {tool} __control__ method returned False")
                    except Exception as e:
                        # Re-raise any exceptions from the __control__ method
                        raise ValueError(f"Error validating tool {tool}: {str(e)}")

    @property
    def retries(self):
        if self.reflection:
            return 5
        else:
            return 1

    @property
    def agent_id(self):
        if self.agent_id_ is None:
            self.agent_id_ = str(uuid.uuid4())
        return self.agent_id_
    
    def do(self, task: Task):
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    self.do_async(task), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.do_async(task))
    
    async def do_async(self, task: Task):
        """Asynchronous version of the do method."""
        return await execute_task_async(self, task, self.debug)
    
    def print_do(self, task: Task):
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                result = asyncio.run_coroutine_threadsafe(
                    self.print_do_async(task), 
                    loop
                ).result()
                return result
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.print_do_async(task))
        
    async def print_do_async(self, task: Task):
        """Asynchronous version of the print_do method."""
        result = await self.do_async(task)
        print(result)
        return result
    
    def parallel_do(self, tasks: List[Task]):
        """Execute multiple tasks in parallel and return their results.
        
        Args:
            tasks: A list of Task objects to execute in parallel
            
        Returns:
            A list of task responses in the same order as the input tasks
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    self.parallel_do_async(tasks), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.parallel_do_async(tasks))
    
    async def parallel_do_async(self, tasks: List[Task]):
        """Asynchronous version of the parallel_do method.
        
        Args:
            tasks: A list of Task objects to execute in parallel
            
        Returns:
            A list of task responses in the same order as the input tasks
        """
        # Create a list of coroutines for each task
        coroutines = [self.do_async(task) for task in tasks]
        
        # Execute all tasks in parallel and return their results
        return await asyncio.gather(*coroutines)
    
    def parallel_print_do(self, tasks: List[Task]):
        """Execute multiple tasks in parallel, print their results, and return them.
        
        Args:
            tasks: A list of Task objects to execute in parallel
            
        Returns:
            A list of task responses in the same order as the input tasks
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    self.parallel_print_do_async(tasks), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.parallel_print_do_async(tasks))
    
    async def parallel_print_do_async(self, tasks: List[Task]):
        """Asynchronous version of the parallel_print_do method.
        
        Args:
            tasks: A list of Task objects to execute in parallel
            
        Returns:
            A list of task responses in the same order as the input tasks
        """
        # Execute all tasks in parallel
        results = await self.parallel_do_async(tasks)
        
        # Print each result
        for result in results:
            print(result)
        
        return results
